﻿using MySql.Data.MySqlClient;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading.Tasks;

namespace MontoyaTech.MySqlPlus
{
    /// <summary>
    /// A set of extensions to help work with a MySqlDataReader.
    /// </summary>
    public static class MySqlDataReaderExtensions
    {
        /// <summary>
        /// Reads a row from a MySqlDataReader and returns an instance of the row.
        /// This function will return null if it can't read the row.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="reader"></param>
        /// <returns></returns>
        public static T Read<T>(this MySqlDataReader reader)
        {
            var row = typeof(T).CreateInstance<T>();

            if (Read<T>(reader, row))
                return row;

            return default;
        }

        /// <summary>
        /// Reads a row from a MySqlDataReader into an existing instance of a row and
        /// returns whether or not it succeeded
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="reader"></param>
        /// <param name="row"></param>
        /// <returns></returns>
        /// <exception cref="Exception"></exception>
        public static bool Read<T>(this MySqlDataReader reader, T row)
        {
            if (!reader.Read())
                return false;

            var rowType = typeof(T);

            var fields = rowType.GetFields();

            if (fields == null || fields.Length == 0)
                throw new Exception("Found no public fields on given row.");

            for (int i = 0; i < fields.Length; i++)
            {
                var column = fields[i].GetCustomAttribute<MySqlColumn>();

                if (column == null)
                    continue;
                else
                    column.ReadValue(row, fields[i], reader);
            }

            return true;
        }

        /// <summary>
        /// Reads a row from a MySqlDataReader and returns whether or not it succeeded.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="reader"></param>
        /// <param name="row"></param>
        /// <returns></returns>
        public static bool Read<T>(this MySqlDataReader reader, out T row)
        {
            row = default(T);

            if (!reader.Read())
                return false;

            var rowType = typeof(T);

            var fields = rowType.GetFields();

            if (fields == null || fields.Length == 0)
                throw new Exception("Found no public fields on given row.");

            row = rowType.CreateInstance<T>();

            for (int i = 0; i < fields.Length; i++)
            {
                var column = fields[i].GetCustomAttribute<MySqlColumn>();

                if (column == null)
                    continue;
                else
                    column.ReadValue(row, fields[i], reader);
            }

            return true;
        }

        /// <summary>
        /// Reads all the rows from a MySqlDataReader.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="reader"></param>
        /// <returns></returns>
        public static List<T> ReadAll<T>(this MySqlDataReader reader)
        {
            var rows = new List<T>();

            ReadAll<T>(reader, rows);

            return rows;
        }

        /// <summary>
        /// Reads all the rows from a MySqlDataReader into an existing list.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="reader"></param>
        /// <param name="rows"></param>
        public static void ReadAll<T>(this MySqlDataReader reader, List<T> rows)
        {
            var rowType = typeof(T);

            var fields = rowType.GetFields();

            if (fields == null || fields.Length == 0)
                throw new Exception("Found no public fields on given row.");

            while (reader.Read())
            {
                var row = rowType.CreateInstance<T>();

                for (int i = 0; i < fields.Length; i++)
                {
                    var column = fields[i].GetCustomAttribute<MySqlColumn>();

                    if (column == null)
                        continue;
                    else
                        column.ReadValue(row, fields[i], reader);
                }

                rows.Add(row);
            }
        }
    }
}
